##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Utility functions and definitions

import os
import re
import stat
import sys

import OFS
from App.Extensions import getObject
from DateTime import DateTime
from OFS.Image import Pdata
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.PythonScripts.PythonScript import PythonScript
from ZODB.utils import p64


try:
    import ZServer  # NOQA: F401 unused import
    HAVE_FTP = True
except ImportError:
    HAVE_FTP = False


_iswin32 = (sys.platform == 'win32' or
            'test' in os.environ.get('_', '') or
            os.environ.get('TOX_ENV_DIR', ''))
_test_read = 1024 * 8
_unknown = '(unknown)'
_marker = []
unc_expr = re.compile(r'(\\\\[^\\]+\\[^\\]+)(.*)')


def _get_content_type(ext, _type_map):
    """_get_content_type"""
    try:
        return _type_map[ext.lower()]
    except (KeyError, AttributeError):
        return (None, None)


def _set_content_type(ob, content_type, data):
    """_set_content_type"""
    if content_type:
        ob.content_type = content_type

    if hasattr(ob, 'content_type') and ob.content_type == 'text/html':
        if content_type == 'text/html':
            return
        data = data.strip().lower()
        if data[:6] != '<html>' and data[:14] != '<!doctype html':
            ob.content_type = 'text/plain'


def _list2typemap(types_list):
    """_list2typemap"""
    if not types_list:
        return
    types_mapping = {}
    for i in types_list:
        if i:
            try:
                e, t = i.split()
                c = ''
            except ValueError:
                try:
                    e, t, c = re.split('[ \t]+', i, 2)
                except ValueError:
                    raise ValueError('Error parsing type map: "%s"' % i)
            types_mapping[e] = (t, c)
    return types_mapping


def _typemap2list(types_mapping):
    """_typemap2list"""
    res = []
    keys = sorted(types_mapping.keys())
    for k in keys:
        v = types_mapping[k]
        if isinstance(v, tuple):
            res.append(' '.join((k, v[0], v[1])))
        else:
            res.append(' '.join((k, v)))
    return res


def _list2iconmap(icons_list):
    """_list2iconmap"""
    if not icons_list:
        return
    icons_mapping = {}
    for i in icons_list:
        if i:
            try:
                k, v = i.split()
            except ValueError:
                raise ValueError("Error parsing icon map: '%s'" % i)
            icons_mapping[k] = v
    return icons_mapping


def _iconmap2list(icons_mapping):
    """_iconmap2list"""
    res = []
    for key in sorted(icons_mapping.keys()):
        res.append(' '.join((key, icons_mapping[key])))
    return res


def _create_ob(id, file, path, _type_map):
    """_create_ob"""
    ob = None
    ext = os.path.splitext(path)[-1]
    t, c = _get_content_type(ext, _type_map)
    if c is not None:
        ob = _create_builtin_ob(c, id, file, path)
        if ob is None:
            ob = _create_ob_from_function(c, id, file, path)
        if ob is None:
            ob = _create_ob_from_factory(c, id, file, path)
    if ob is None:
        ob = _wrap_ob(_create_File(id, file), path)
    file.seek(0)
    ob.__doc__ = 'LocalFile'
    _set_content_type(ob, t, file.read(_test_read))
    return ob


def _create_DTMLMethod(id, file):
    """_create_DTMLMethod"""
    return OFS.DTMLMethod.DTMLMethod(file.read().decode(), __name__=id)


def _create_DTMLDocument(id, file):
    """_create_DTMLDocument"""
    return OFS.DTMLDocument.DTMLDocument(file.read().decode(), __name__=id)


def _create_Image(id, file):
    """_create_Image"""
    # return StreamingImage(id, '', file)
    return OFS.Image.Image(id, '', file)


def _create_File(id, file):
    """_create_File"""
    # return StreamingFile(id, '', file)
    return OFS.Image.File(id, '', file)


def _create_ZPT(id, file):
    """_create_ZPT"""
    return ZopePageTemplate(id, file, content_type='text/html')


def _create_PythonScript(id, file):
    """_create_PythonScript"""
    ob = PythonScript(id)
    ob.write(file.read())
    return ob


_builtin_create = {
    'PythonScript': _create_PythonScript,
    'DTMLMethod': _create_DTMLMethod,
    'DTMLDocument': _create_DTMLDocument,
    'Image': _create_Image,
    'File': _create_File,
    'PageTemplate': _create_ZPT,  # ***SmileyChris
}


def _create_builtin_ob(c, id, file, path):
    try:
        f = _builtin_create[c]
        return _wrap_ob(f(id, file), path)
    except Exception:
        pass


def _create_ob_from_function(c, id, file, path):
    try:
        i = c.rindex('.')
        m, c = c[:i], c[i+1:]
        m = __import__(m, globals(), locals(), (c,))
        c = getattr(m, c)
        f = getattr(c, 'createSelf').__func__
        if f.__code__.co_varnames == ('id', 'file'):
            return _wrap_ob(f(id, file), path)
    except Exception:
        pass


def _create_ob_from_factory(c, id, file, path):
    try:
        i = c.rindex('.')
        m, c = c[:i], c[i+1:]
        c = getObject(m, c)
        f = c()
        ob = _wrap_ob(f(id, file), path)
        ob.__factory = f
        return ob
    except Exception:
        pass


class Wrapper:
    """Mix-in class used to save object changes."""
    _local_path = None
    # Create a global and persistent lock table
    _dav_writelocks = {}
    # The object itself is not persistent and cannot be stored
    _p_changed = 0

    def bobobase_modification_time(self):
        """ bobobase_modification_time """
        t = os.stat(self._local_path)[stat.ST_MTIME]
        return DateTime(t)

    def __repr__(self):
        """ __repr__ """
        c = self.__class__.__bases__[-1].__name__
        return '<%s ObjectWrapper instance at %8X>' % (c, id(self))

    def wl_lockmapping(self, killinvalids=0, create=0):
        """ Overwrite the default method of LockableItem """
        locks = self._dav_writelocks.get(self._local_path)
        if locks is None:
            locks = {}
            if create:
                # Store it in persistent lock table
                self._dav_writelocks[self._local_path] = locks
        elif killinvalids:
            # Delete invalid locks
            for token, lock in locks.items():
                if not lock.isValid():
                    del locks[token]
            if not locks and not create:
                # Remove empty lock table
                del self._dav_writelocks[self._local_path]
        return locks


_wrapper_method = '''def %(name)s %(arglist)s:
    """Wrapper for the %(name)s method."""
    # r = apply(self.__class__.__bases__[-1].%(name)s, %(baseargs)s)
    r = self.__class__.__bases__[-1].%(name)s(*%(baseargs)s)
    try: _save_ob(self, self._local_path)
    except ValueError: pass
    return r
'''

_wrappers = {}


def _get_wrapper(c):
    try:
        return _wrappers[c]
    except KeyError:
        class ObjectWrapper(Wrapper, c):
            pass

        _wrap_method(ObjectWrapper, 'manage_edit')
        _wrap_method(ObjectWrapper, 'manage_upload')
        _wrap_method(ObjectWrapper, 'pt_edit')  # PT ZMI editing
        _wrap_method(ObjectWrapper, 'write')  # PythonScript
        _wrap_method(ObjectWrapper, 'ZBindingsHTML_editAction')  # PythonScript
        _wrap_method(ObjectWrapper, 'PUT')
        # Remove management options that cannot be used
        manage_options = []
        for opt in ObjectWrapper.manage_options:
            if opt['label'] not in ('Properties', 'Security', 'Undo',
                                    'Ownership', 'Interfaces', 'Proxy',
                                    'History'):
                manage_options.append(opt)
        ObjectWrapper.manage_options = tuple(manage_options)
        _wrappers[c] = ObjectWrapper
        return ObjectWrapper


def _wrap_method(c, name):
    try:
        m = getattr(c.__bases__[-1], name)
    except AttributeError:
        return
    try:
        f = m.__func__
    except AttributeError:
        f = m
    a = list(f.__code__.co_varnames)[:f.__code__.co_argcount]
    d = f.__defaults__ or ()
    arglist = []
    baseargs = []
    while (len(a) > len(d)):
        arglist.append(a[0])
        baseargs.append(a[0])
        del a[0]
    for i in range(len(a)):
        arglist.append('%s=%s' % (a[i], repr(d[i])))
        baseargs.append(a[i])
    arglist = '(' + ','.join(arglist) + ')'
    baseargs = '(' + ','.join(baseargs) + ')'
    d = {}
    # exec _wrapper_method % vars() in globals(), d
    exec(_wrapper_method % vars(), globals(), d)
    setattr(c, name, d[name])


def _wrap_ob(ob, path):
    c = ob.__class__
    n = ob.__name__
    if hasattr(c, '__basicnew__'):
        c = _get_wrapper(c)
        d = ob.__dict__
        ob = c.__basicnew__()
        ob.__dict__.update(d)
    else:
        c = _get_wrapper(c)
        ob.__class__ = c
    ob._local_path = path
    ob.__name__ = n
    ob._p_oid = path
    return ob


def _save_DTML(ob, path):
    f = open(path, 'w')
    try:
        f.write(ob.read())
    finally:
        f.close()


def _save_File(ob, path):
    with open(path, 'wb') as fp:
        if isinstance(ob.data, Pdata):
            fp.write(bytes(ob.data))
        else:
            fp.write(ob.data)


def _save_Folder(ob, path):
    os.mkdir(path)


_builtin_save = {
    'Script (Python)': _save_DTML,
    'DTML Method': _save_DTML,
    'DTML Document': _save_DTML,
    'Image': _save_File,
    'File': _save_File,
    'Folder': _save_Folder,
    'Page Template': _save_DTML,  # ***SmileyChris
}


def _save_builtin_ob(ob, path):
    try:
        f = _builtin_save[ob.meta_type]
        f(ob, path)
        return 1
    except KeyError:
        pass


def _save_ob_with_function(ob, path):
    try:
        ob.saveSelf(path)
        return 1
    except Exception:
        pass


def _save_ob_with_factory(ob, path):
    try:
        ob.__factory.save(ob, path)
        return 1
    except Exception:
        pass


def _save_ob(ob, path):
    s = _save_builtin_ob(ob, path)
    if not s:
        s = _save_ob_with_function(ob, path)
    if not s:
        s = _save_ob_with_factory(ob, path)
    if not s:
        raise TypeError("Cannot save files of type '%s'." % ob.meta_type)


def _set_timestamp(ob, path):
    t = os.stat(path)[stat.ST_MTIME]
    ob._p_serial = p64(t)


_marker = []


def valid_id(id):
    if id == os.curdir or id == os.pardir or id[0] == '_':
        return 0
    return 1


bad_id = re.compile('[^a-zA-Z0-9-_~,. ]').search


def sanity_check(c, ob):
    # This is called on cut/paste operations to make sure that
    # an object is not cut and pasted into itself or one of its
    # subobjects, which is an undefined situation.
    dest = c.basepath
    src = ob._local_path
    if dest[:len(src)] != src:
        return 1



from io import BytesIO
from io import StringIO

import six

from OFS.Image import File
from OFS.Image import Image
from OFS.Image import Pdata
from ZPublisher.HTTPRequest import FileUpload


BUFFER_SIZE = 1 << 16


def _read_data(self, file):
    # We do not want to load the whole file into memory, so just
    # get the file size and return a faked Pdata object.
    if isinstance(file, (bytes, str)):
        size = len(file)
        if size < BUFFER_SIZE:
            return file, size
        # Big string: cut it into smaller chunks
        if isinstance(file, bytes):
            file = BytesIO(file)
        else:
            file = StringIO(file)

    if isinstance(file, FileUpload) and not file:
        raise ValueError('File not specified')

    if hasattr(file, '__class__') and file.__class__ in (Pdata, Sdata):
        size = len(file)
        return file, size

    pos = file.tell()
    file.seek(0, 2)
    size = file.tell()
    file.seek(pos, 0)
    return Sdata(file, size), size


class StreamingFile(File):
    """ Wrapper around OFS.Image.File """
    _read_data = _read_data


class StreamingImage(Image):
    """ Wrapper around OFS.Image.Image """
    _read_data = _read_data


class Sdata(Pdata):
    """ Streaming wrapper for possibly large data """
    # Imitates OFS.Image.Pdata
    # Make it a subclass of Pdata to be conform with ExternalEditor

    _p_changed = 0

    def __init__(self, file, fsize, _offset=0):
        self.file = file
        self.fsize = fsize
        self.offset = _offset

    def __getitem__(self, key):
        size = min(BUFFER_SIZE, len(self))
        if isinstance(key, int):
            if key > size:
                return b''

            self.file.seek(self.offset+key, 0)
            return self.file.read(1)

    @property
    def data(self):
        return self[0:BUFFER_SIZE]

    @property
    def next(self):
        offset = self.offset + BUFFER_SIZE
        if offset < self.fsize:
            return Sdata(self.file, self.fsize, offset)

    def __len__(self):
        return self.fsize - self.offset

    def __bytes__(self):
        self.file.seek(self.offset, 0)
        return self.file.read()

    if six.PY2:
        def __getslice__(self, i, j):
            size = min(BUFFER_SIZE, len(self))
            if i < 0:
                i = max(i + size, 0)
            j = min(j, size)
            if j < 0:
                j = max(j + size, 0)
            if i >= j:
                return ''
            self.file.seek(self.offset+i, 0)
            return self.file.read(j-i)

        __str__ = __bytes__




class XMLDocumentFactory:

    def __call__(self, id, file):
        try:
            from Products.ParsedXML.ParsedXML import ParsedXML
        except ImportError:
            return
        ob = ParsedXML(id=id, XMLstring=file.read(), namespaces=1,
                       contentType='text/xml')
        return ob

    def save(self, ob, path):
        f = open(path, 'w')
        f.write(ob.index_html())
        f.close()



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Tests for the LocalFS class

import unittest

from Products.PageTemplates.ZopePageTemplate import manage_addPageTemplate
from Testing.makerequest import makerequest

from ..LocalFS import _iswin32
from .helpers import ADMIN_USER
from .helpers import LOCALFS_ROOT
from .helpers import FilesystemTestSupport
from .helpers import FunctionalTestCase


class LocalFSTests(unittest.TestCase, FilesystemTestSupport):

    def _getTargetClass(self):
        from ..LocalFS import LocalFS

        return LocalFS

    def _makeOne(self, *args):
        return self._getTargetClass()(*args)

    def _makeSimple(self):
        klass = self._getTargetClass()
        return klass('localfs', 'LocalFS Title', LOCALFS_ROOT, 'user', 'pw')

    def tearDown(self):
        self.cleanup_files()
        super(LocalFSTests, self).tearDown()

    def test_instantiation(self):
        lfs = self._makeSimple()
        self.assertEqual(lfs.getId(), 'localfs')
        self.assertEqual(lfs.title, 'LocalFS Title')
        self.assertEqual(lfs.basepath, LOCALFS_ROOT)

        # Only on Windows
        if _iswin32:
            self.assertEqual(lfs.username, 'user')
            self.assertEqual(lfs._password, 'pw')

    def test_factory(self):
        from OFS.Folder import Folder

        from ..LocalFS import manage_addLocalFS

        folder = Folder('test')

        manage_addLocalFS(folder, 'lfs', 'Local FS', LOCALFS_ROOT)

        self.assertEqual(folder.lfs.getId(), 'lfs')
        self.assertEqual(folder.lfs.title, 'Local FS')
        self.assertEqual(folder.lfs.basepath, LOCALFS_ROOT)

        # Only on Windows
        if _iswin32:
            self.assertEqual(folder.lfs.username, None)
            self.assertEqual(folder.lfs._password, '')

    def test_hasDefaultDocument(self):
        lfs = self._makeSimple()
        lfs.default_document = 'index.html default.html'

        self.assertFalse(lfs.hasDefaultDocument())

        manage_addPageTemplate(lfs, 'index.html', b'<html>')
        self.assertTrue(lfs.hasDefaultDocument())


class LocalFSFunctionalTests(FunctionalTestCase):

    def test_manage_editProperties(self):
        self.login(ADMIN_USER)
        lfs = makerequest(self.folder.localfs)
        req = lfs.REQUEST

        req.form['title'] = 'New Title'
        req.form['basepath'] = '/tmp'
        req.form['username'] = 'user1'
        req.form['password'] = 'pass1'
        req.form['default_document'] = 'my.html index.jsp'
        req.form['catalog'] = 1
        req.form['type_map'] = ['',
                                '.dtml text/html DTMLMethod',
                                '.gif image/gif Image',
                                '.foo Foobar']
        req.form['icon_map'] = ['', 'directory dir.gif', 'foobar foo.gif']
        req.form['file_filter'] = 'None'

        lfs.manage_editProperties(req)
        self.assertEqual(lfs.title, 'New Title')
        self.assertEqual(lfs.basepath, '/tmp')
        self.assertEqual(lfs.default_document, 'my.html index.jsp')
        self.assertEqual(lfs.catalog, 1)
        self.assertEqual(lfs._type_map,
                         {'.dtml': ('text/html', 'DTMLMethod'),
                          '.gif': ('image/gif', 'Image'),
                          '.foo': ('Foobar', '')})
        self.assertEqual(lfs._icon_map,
                         {'directory': 'dir.gif', 'foobar': 'foo.gif'})
        self.assertIsNone(lfs.file_filter)
        self.assertTrue(lfs.isPrincipiaFolderish)
        self.assertTrue(lfs.catalog)

        if _iswin32:  # Only on Windows
            self.assertEqual(lfs.username, 'user1')
            self.assertEqual(lfs._password, 'pass1')
            self.assertEqual(lfs.password, '')  # Gets reset automatically

        # Some special cases
        req.form['file_filter'] = 'myfilter'
        lfs.manage_editProperties(req)
        self.assertEqual(lfs.file_filter, 'myfilter')
        req.form['file_filter'] = ' '
        lfs.manage_editProperties(req)
        self.assertIsNone(lfs.file_filter)

    def test_manage_changeProperties(self):
        self.login(ADMIN_USER)
        lfs = makerequest(self.folder.localfs)

        new_type_map = ['',
                        '.dtml text/html DTMLMethod',
                        '.gif image/gif Image',
                        '.foo Foobar']
        new_icon_map = ['', 'directory dir.gif', 'foobar foo.gif']

        lfs.manage_changeProperties(REQUEST=None,
                                    title='New Title',
                                    basepath='/tmp',
                                    username='user1',
                                    password='pass1',
                                    default_document='my.html index.jsp',
                                    catalog=1,
                                    type_map=new_type_map,
                                    icon_map=new_icon_map,
                                    file_filter='')

        self.assertEqual(lfs.title, 'New Title')
        self.assertEqual(lfs.basepath, '/tmp')
        self.assertEqual(lfs.default_document, 'my.html index.jsp')
        self.assertEqual(lfs.catalog, 1)
        self.assertEqual(lfs._type_map,
                         {'.dtml': ('text/html', 'DTMLMethod'),
                          '.gif': ('image/gif', 'Image'),
                          '.foo': ('Foobar', '')})
        self.assertEqual(lfs._icon_map,
                         {'directory': 'dir.gif', 'foobar': 'foo.gif'})
        self.assertFalse(lfs.file_filter)
        self.assertTrue(lfs.isPrincipiaFolderish)
        self.assertTrue(lfs.catalog)

        if _iswin32:  # Only on Windows
            self.assertEqual(lfs.username, 'user1')
            self.assertEqual(lfs._password, 'pass1')
            self.assertEqual(lfs.password, '')  # Gets reset automatically



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Tests for the StreamingFile class

import tempfile
import unittest

from .helpers import LOCALFS_ROOT


class StreamingFileTests(unittest.TestCase):

    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile(dir=LOCALFS_ROOT,
                                                    delete=True)
        self.tempfile.write(b'.'*40000)

    def tearDown(self):
        self.tempfile.close()

    def _getTargetClass(self):
        from ..StreamingFile import StreamingFile

        return StreamingFile

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def _makeSimple(self, data=b'data'):
        return self._makeOne('fileid', '', data)

    def test_read(self):
        from ..StreamingFile import BUFFER_SIZE
        from ..StreamingFile import Sdata

        ob = self._makeSimple()

        # small file
        data = b'data'
        ob = self._makeSimple(data=data)
        self.assertEqual(ob.getId(), 'fileid')
        self.assertEqual(ob.data, data)
        self.assertEqual(ob._read_data(file=data), (b'data', 4))

        # Large data
        data = b'.' * BUFFER_SIZE
        ob = self._makeSimple(data=data)
        self.assertEqual(bytes(ob.data), data)
        stream_data, size = ob._read_data(ob.data)
        self.assertIsInstance(stream_data, Sdata)
        self.assertEqual(size, BUFFER_SIZE)
        self.assertEqual(bytes(stream_data), data)

        # Read a single byte
        self.assertEqual(stream_data[BUFFER_SIZE-1], b'.')

        # Try to read beyond chunk size
        self.assertEqual(stream_data[BUFFER_SIZE+1], b'')



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Tests for the utility functions

import os
import tempfile
import unittest

from .helpers import LOCALFS_ROOT
from .helpers import FakeContent


class UtilsTests(unittest.TestCase):

    def test__get_content_type(self):
        from ..LocalFS import _types
        from ..utils import _get_content_type

        # Valid input
        self.assertEqual(_get_content_type('.gif', _types),
                         ('image/gif', 'Image'))
        self.assertEqual(_get_content_type('.GIF', _types),
                         ('image/gif', 'Image'))

        # Invalid input or unknown type
        self.assertEqual(_get_content_type('.foo', _types), (None, None))
        self.assertEqual(_get_content_type(None, _types), (None, None))

    def test_set_content_type(self):
        from ..utils import _set_content_type

        data = b''
        ob = FakeContent()
        _set_content_type(ob, 'image/gif', data)
        self.assertEqual(ob.content_type, 'image/gif')

        # If the object already has a content_type attribute and a non-true
        # value is passed in then it will not be changed
        _set_content_type(ob, None, data)
        self.assertEqual(ob.content_type, 'image/gif')
        _set_content_type(ob, '', data)
        self.assertEqual(ob.content_type, 'image/gif')

        # (Weird) special handling for text/html
        _set_content_type(ob, 'text/html', data)
        self.assertEqual(ob.content_type, 'text/html')
        _set_content_type(ob, None, data)
        self.assertEqual(ob.content_type, 'text/plain')

        data = '<html>...'
        _set_content_type(ob, 'text/html', data)
        self.assertEqual(ob.content_type, 'text/html')
        _set_content_type(ob, None, data)
        self.assertEqual(ob.content_type, 'text/html')

        data = '<!doctype html>...'
        _set_content_type(ob, 'text/html', data)
        self.assertEqual(ob.content_type, 'text/html')
        _set_content_type(ob, None, data)
        self.assertEqual(ob.content_type, 'text/html')

    def test__list2typemap(self):
        from ..utils import _list2typemap

        t_list = ['',
                  '.dtml text/html DTMLMethod',
                  '.gif image/gif Image',
                  '.foo Foobar']
        self.assertEqual(_list2typemap(t_list),
                         {'.dtml': ('text/html', 'DTMLMethod'),
                          '.gif': ('image/gif', 'Image'),
                          '.foo': ('Foobar', '')})

        # Invalid input
        self.assertIsNone(_list2typemap(None))
        self.assertIsNone(_list2typemap(''))
        self.assertIsNone(_list2typemap([]))

    def test__typemap2list(self):
        from ..LocalFS import _types
        from ..utils import _typemap2list

        t_list = _typemap2list(_types)
        self.assertIsInstance(t_list, list)
        self.assertEqual(set(_types.keys()),
                         set([x.split()[0] for x in t_list]))

        t_map = {'.dtml': ('text/html', 'DTMLMethod'),
                 '.foo': 'Foobar'}
        self.assertEqual(_typemap2list(t_map),
                         ['.dtml text/html DTMLMethod', '.foo Foobar'])

    def test__list2iconmap(self):
        from ..utils import _list2iconmap

        i_list = ['', 'directory dir.gif', 'application binary.gif']
        self.assertEqual(_list2iconmap(i_list),
                         {'directory': 'dir.gif', 'application': 'binary.gif'})

        # Invalid input
        self.assertIsNone(_list2iconmap(None))
        self.assertIsNone(_list2iconmap(''))
        self.assertIsNone(_list2iconmap([]))
        with self.assertRaises(ValueError):
            _list2iconmap(['invalid'])

    def test__iconmap2list(self):
        from ..LocalFS import _icons
        from ..utils import _iconmap2list

        i_list = _iconmap2list(_icons)
        self.assertIsInstance(i_list, list)
        self.assertEqual(set(_icons.keys()),
                         set([x.split()[0] for x in i_list]))

    def test__set_timestamp(self):
        from ..utils import _set_timestamp

        ob = FakeContent()
        with tempfile.NamedTemporaryFile(dir=LOCALFS_ROOT) as fp:
            _set_timestamp(ob, fp.name)
        self.assertIsInstance(ob._p_serial, bytes)

    def test_valid_id(self):
        from ..utils import valid_id

        self.assertFalse(valid_id(os.curdir))
        self.assertFalse(valid_id(os.pardir))
        self.assertFalse(valid_id('_secret'))
        self.assertTrue(valid_id('foobar'))

    def test_sanity_check(self):
        from ..utils import sanity_check

        copy_source = FakeContent()
        copy_source._local_path = os.path.join(LOCALFS_ROOT, 'foo.txt')
        copy_target = FakeContent()

        copy_target.basepath = os.path.join(LOCALFS_ROOT, 'bar.txt')
        self.assertTrue(sanity_check(copy_target, copy_source))

        copy_target.basepath = os.path.join(LOCALFS_ROOT, 'foo.txt')
        self.assertFalse(sanity_check(copy_target, copy_source))



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
import os

from App.ImageFile import ImageFile

from . import LocalFS


misc_ = {}
www_files = os.listdir(os.path.join(os.path.dirname(__file__), 'www'))
icons = [f for f in www_files if f[-4:] == '.gif']

for icon in icons:
    misc_[icon] = ImageFile('www/%s' % icon, globals())


def initialize(context):
    context.registerClass(
        LocalFS.LocalFS,
        constructors=(LocalFS.manage_addLocalFSForm,
                      LocalFS.manage_addLocalFS),
        icon='www/fs.gif',
        )



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
"""Local File System product"""
__version__ = '1.12'
__doc__ = """Local File System product"""

import glob
import marshal  # sbk
import os
import stat

import six
from six.moves.urllib.parse import quote

import AccessControl
import Acquisition
import App
import OFS
import Persistence
from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.Common import absattr
from App.FactoryDispatcher import ProductDispatcher
from App.special_dtml import HTMLFile
from DateTime import DateTime
from OFS.CopySupport import CopyError
from OFS.CopySupport import _cb_decode
from OFS.CopySupport import _cb_encode
from OFS.role import RoleManager
from webdav.NullResource import LockNullResource
from zExceptions import BadRequest
from zExceptions import Forbidden
from zExceptions import NotFound
from zExceptions import Unauthorized

from .exceptions import DeleteError
from .exceptions import RenameError
from .exceptions import UploadError
from .utils import HAVE_FTP
from .utils import _create_ob
from .utils import _get_content_type
from .utils import _iconmap2list
from .utils import _iswin32
from .utils import _list2iconmap
from .utils import _list2typemap
from .utils import _marker
from .utils import _save_ob
from .utils import _set_content_type
from .utils import _set_timestamp
from .utils import _test_read
from .utils import _typemap2list
from .utils import _unknown
from .utils import bad_id
from .utils import sanity_check
from .utils import unc_expr
from .utils import valid_id


if (_iswin32):
    try:
        import win32wnet
    except ImportError:
        pass


############################################################################
# The process of determining the content-type involves the following steps.
# This same process is used in LocalDirectory._getOb and LocalFile._getType.
#
# 1. Call _get_content_type which tries to look up the content-type
#    in the type map.
# 2. If that fails then create a file object with the first _test_read
#    bytes of data and see what content-type Zope determines.
#    If we can't read the file then assign 'application/octet-stream'.
# 3. If we found a content-type then assign it to the object in
#    _set_content_type. This overrides the type assigned by Zope.
# 4. Try to see if Zope has assigned 'text/html' to a file that it
#    shouldn't have and change the type back to 'text/plain'. This
#    also happens in _set_content_type.
############################################################################


_types = {
    '.py': ('text/x-python', 'PythonScript'),
    '.html': ('text/html', 'DTMLDocument'),
    '.htm': ('text/html', 'DTMLDocument'),
    '.dtml': ('text/html', 'DTMLMethod'),
    '.gif': ('image/gif', 'Image'),
    '.jpg': ('image/jpeg', 'Image'),
    '.jpeg': ('image/jpeg', 'Image'),
    '.png': ('image/png', 'Image'),
    '.pt': ('text/html', 'PageTemplate'),  # ***SmileyChris
    '.zpt': ('text/html', 'PageTemplate'),  # ***SmileyChris
    '.ra': ('audio/vnd.rn-realaudio', ''),
    '.rv': ('video/vnd.rn-realvideo', ''),
    '.rm': ('application/vnd.rn-realmedia', ''),
    '.rp': ('image/vnd.rn-realpix', ''),
    '.rt': ('text/vnd.rn-realtext', ''),
    '.smi': ('application/smil', ''),
    '.swf': ('application/x-shockwave-flash', ''),
    '.stx': (
        'text/html',
        'Products.StructuredDocument.StructuredDocument.StructuredDocument'),
    '.xml': ('text/xml', 'LocalFS.Factory.XMLDocumentFactory'),
}

_icons = {
    'directory': 'dir.gif',
    'application': 'binary.gif',
    'application/mac-binhex40': 'compressed.gif',
    'application/octet-stream': 'binary.gif',
    'application/pdf': 'layout.gif',
    'application/postscript': 'ps.gif',
    'application/smil': 'layout.gif',
    'application/vnd.rn-realmedia': 'movie.gif',
    'application/x-dvi': 'dvi.gif',
    'application/x-gtar': 'tar.gif',
    'application/x-gzip': 'compressed.gif',
    'application/x-shockwave-flash': 'image2.gif',
    'application/x-tar': 'tar.gif',
    'application/x-tex': 'tex.gif',
    'application/zip': 'compressed.gif',
    'audio': 'sound1.gif',
    'audio/mpeg': 'sound1.gif',
    'audio/vnd.rn-realaudio': 'sound1.gif',
    'audio/x-wav': 'sound1.gif',
    'image': 'image2.gif',
    'image/gif': 'image2.gif',
    'image/jpeg': 'image2.gif',
    'image/png': 'image2.gif',
    'image/vnd.rn-realpix': 'image2.gif',
    'text': 'text.gif',
    'text/html': 'layout.gif',
    'text/plain': 'text.gif',
    'text/x-python': 'p.gif',
    'video': 'movie.gif',
    'video/mpeg': 'movie.gif',
    'video/quicktime': 'movie.gif',
    'video/vnd.rn-realvideo': 'movie.gif',
}
_icon_base = 'misc_/LocalFS/'
for k, v in _icons.items():
    _icons[k] = _icon_base + v


class LocalDirectory(OFS.CopySupport.CopyContainer, App.Management.Navigation,
                     OFS.SimpleItem.Item, Acquisition.Implicit):

    """Object representing a directory in the local file system."""

    meta_type = 'Local Directory'
    zmi_icon = 'far fa-folder text-danger'

    # Set to True and the ZMI becomes too sluggish
    isPrincipiaFolderish = 0
    manage_addProduct = ProductDispatcher()

    manage_main = HTMLFile('dtml/main', globals())
    manage_uploadForm = HTMLFile('dtml/methodUpload', globals())

    manage_options = (
        {'label': 'Contents', 'action': 'manage_main'},
        {'label': 'Upload', 'action': 'manage_uploadForm'},
        {'label': 'View', 'action': ''},
        )

    icon = 'misc_/OFSP/Folder_icon.gif'

    security = AccessControl.ClassSecurityInfo()  # sbk

    security.declareProtected('FTP access', 'manage_FTPstat',  # NOQA
                              'manage_FTPget', 'manage_FTPlist')

    def __init__(self, id, basepath, root, catalog, _type_map,
                 _icon_map, file_filter=None):
        """LocalDirectory __init__"""
        self.id = id
        self.basepath = self._local_path = basepath
        self.root = root
        self.catalog = catalog
        self._type_map = _type_map
        self._icon_map = _icon_map
        self.file_filter = file_filter

    def __bobo_traverse__(self, REQUEST, name):
        """ bobo_traverse """
        method = REQUEST.get('REQUEST_METHOD', 'GET').upper()
        try:
            # FTP - PUT
            if method not in ('GET', 'POST', 'HEAD'):
                return LockNullResource(name).__of__(self)
        except Exception:
            pass
        try:
            return self._safe_getOb(name)
        except Exception:
            pass
        try:
            return getattr(self, name)
        except AttributeError:
            pass

    def __getitem__(self, i):
        if isinstance(i, (str, bytes)):
            return self._safe_getOb(i)
        else:
            raise TypeError('index must be a string')

    def __getattr__(self, attr):
        try:
            return self._safe_getOb(attr)
        except NotFound:
            raise AttributeError(attr)

    def _getpath(self, id):
        return os.path.join(self.basepath, id)

    def _getfileob(self, id, spec=None):
        if spec is None:
            spec = self.file_filter
        path = self._getpath(id)
        return LocalFile(self, id, path, spec)

    def _ids(self, spec=None):
        if spec is None:
            spec = self.file_filter
        try:
            ids = os.listdir(self.basepath)
        except PermissionError:
            raise Forbidden('Sorry, you do not have permission to read '
                            'the requested directory.')

        if (spec is not None):
            try:
                if isinstance(spec, str):
                    spec = spec.split(' ')
                curdir = os.getcwd()
                os.chdir(self.basepath)
                candidates = []
                for id in ids:
                    if os.path.isdir(id) and '*/' in spec or '*\\' in spec:
                        candidates.append(id)
                for patt in spec:
                    names = glob.glob(patt)
                    for id in names:
                        if id[-1] == os.sep:
                            id = id[:-1]
                        if (id not in candidates):
                            candidates.append(id)
                ids = candidates
            finally:
                os.chdir(curdir)
        ids = sorted(filter(valid_id, ids))
        return ids

    def _safe_getOb(self, name, default=_marker):
        return self._getOb(name, default)

    def _getOb(self, id, default=_marker):
        if (id in (os.curdir, os.pardir)):
            raise ValueError(id)
        ob = None
        path = self._getpath(id)
        if (os.path.isdir(path)):
            ob = LocalDirectory(id, path, self.root or self, self.catalog,
                                self._type_map, self._icon_map)
        elif (os.path.isfile(path)):
            try:
                with open(path, 'rb') as f:
                    ob = _create_ob(id, f, path, self._type_map)
            except PermissionError:
                raise Forbidden('Sorry, you do not have permission to read '
                                'the requested resource "%s".' % id)

        if (ob is None):
            if default is _marker:
                raise AttributeError(id)
            return default
        _set_timestamp(ob, path)
        ob._p_jar = self._p_jar
        return ob.__of__(self)

    def _setObject(self, id, object, roles=None, user=None):
        if getattr(object, '__locknull_resource__', 0):
            self._checkId(id, 1)
            return id
        else:
            self._checkId(id)
        self._safe_setOb(id, object)
        return id

    def _delObject(self, id, dp=1):
        self._delOb(id)

    def _checkId(self, id, allow_dup=0):
        # If allow_dup is false, an error will be raised if an object
        # with the given id already exists. If allow_dup is true,
        # only check that the id string contains no illegal chars.
        if not id:
            raise BadRequest('No id was specified')
        if bad_id(id):
            raise BadRequest(
                'The id %s contains characters illegal in filenames.' % id)
        if id[0] == '_':
            raise BadRequest(
                'The id %s  is invalid - it begins with an underscore.' % id)
        if not allow_dup:
            path = self._getpath(id)
            if os.path.exists(path):
                raise BadRequest(
                    'The id %s is invalid - it is already in use.' % id)

    def _safe_setOb(self, id, ob):
        try:
            self._setOb(id, ob)
        except PermissionError:
            raise Forbidden('Sorry, you do not have permission to write '
                            'to this directory.')

    def _setOb(self, id, ob):
        if not hasattr(ob, 'meta_type'):
            raise BadRequest('Unknown object type.')
        path = self._getpath(id)
        try:
            _save_ob(ob, path)
        except TypeError:
            raise BadRequest(
                'Cannot add objects of type "%s" to local directories.'
                % ob.meta_type)

    def _delOb(self, id):
        path = self._getpath(id)
        try:
            if os.path.isdir(path):
                t = 'directory'
                os.rmdir(path)
            else:
                t = 'file'
                os.unlink(path)
        except PermissionError:
            raise Forbidden('Sorry, you do not have permission to delete '
                            'the requested %s ("%s").' % (t, id))
        except OSError:
            if t == 'directory' and os.listdir(path):
                raise DeleteError('The directory "%s" is not empty.' % id)
            raise

    def _copyOb(self, id, ob):
        self._setObject(id, ob)

    def _moveOb(self, id, ob):
        src = ob._local_path
        dest = self._getpath(id)
        try:
            os.rename(src, dest)
        except PermissionError:
            raise Forbidden('Sorry, you do not have permission to write '
                            'to this directory.')

    def _verifyObjectPaste(self, ob, REQUEST):
        pass

    def _write_file(self, file, path):
        try:
            if isinstance(file, bytes):
                outfile = open(path, 'wb')
                outfile.write(file)
                outfile.close()
            else:
                blocksize = 8*1024
                outfile = open(path, 'wb')
                data = file.read(blocksize)
                while data:
                    outfile.write(data)
                    data = file.read(blocksize)
                outfile.close()
        except PermissionError:
            raise Forbidden('Sorry, you do not have permission to write '
                            'to this directory.')

    def manage_createDirectory(self, path, action='manage_workspace',
                               REQUEST=None):
        """Create a new directory relative to this directory."""
        parts = [p for p in os.path.split(path) if p not in ('.', '..')]
        path = os.path.join(*parts)
        fullpath = os.path.join(self.basepath, path)
        if os.path.exists(fullpath):
            if REQUEST is not None:
                REQUEST.set('manage_tabs_message',
                            'Directory %s already exists.' % path)
                return self.manage_uploadForm(self, REQUEST)

        else:
            try:
                os.makedirs(fullpath)
            except PermissionError:
                msg = 'You do not have permission to write to this directory.'
                if REQUEST is not None:
                    REQUEST.set('manage_tabs_message', msg)
                    return self.manage_uploadForm(self, REQUEST)
                else:
                    raise Forbidden(msg)

            if REQUEST is not None:
                msg = 'Directory %s has been created.' % path
                REQUEST.set('manage_tabs_message', msg)
                return self.manage_main(self, REQUEST, update_menu=1)

    def index_html(self, REQUEST, RESPONSE):
        """ Show the default page if it exists, otherwise acquire """
        default_document = self.defaultDocument()
        if default_document is None:
            return aq_parent(aq_inner(self)).index_html(REQUEST, RESPONSE)
        return default_document(REQUEST, RESPONSE)

    def manage_upload(self, file, id='', action='manage_workspace',
                      REQUEST=None):
        """Upload a file to the local file system. The 'file' parameter
        is a FileUpload instance representing the uploaded file."""
        if hasattr(file, 'filename'):
            filename = file.filename
        else:
            filename = file.name
        if not id:
            # Try to determine the filename without any path.
            # First check for a UNIX full path. There will be no UNIX path
            # separators in a Windows path.
            if '/' in filename:
                id = filename[filename.rfind('/')+1:]
            # Next check for Window separators. If there are no UNIX path
            # separators then it's probably a Windows path and not a random
            # relative UNIX path which happens to have backslashes in it.
            # Lets hope this never happens, anyway. ;)
            elif '\\' in filename:
                id = filename[filename.rfind('\\')+1:]
            # Not sure if we'll ever get a Mac path, but here goes...
            elif ':' in filename:
                id = filename[filename.rfind(':')+1:]
            # Else we have a filename with no path components so let's use
            # that for the id.
            else:
                id = filename

        try:
            self._checkId(id, 1)
        except BadRequest as exc:
            if REQUEST is not None:
                REQUEST.set('manage_tabs_message', str(exc))
                return self.manage_uploadForm(self, REQUEST)
            else:
                raise UploadError(str(exc))

        path = self._getpath(id)
        if os.path.exists(path):
            self.manage_overwrite(file, path, REQUEST)
        else:
            self._write_file(file, path)

        if REQUEST:
            REQUEST.set('manage_tabs_message',
                        'File %s has been uploaded.' % id)
            return self.manage_main(self, REQUEST, update_menu=1)

    def manage_overwrite(self, file, path, REQUEST=None):
        """Overwrite a local file."""
        user = getSecurityManager().getUser()

        if user is None or \
           not user.has_permission('Overwrite local files', self):
            raise Unauthorized(
                    'Sorry, you are not authorized to overwrite files.')
        self._write_file(file, path)

    def manage_renameObject(self, id, new_id, REQUEST=None):
        """Rename a file or subdirectory."""
        error_msg = ''
        try:
            self._checkId(new_id)
            f = self._getpath(id)
            t = self._getpath(new_id)
            os.rename(f, t)
        except BadRequest as exc:
            error_msg = str(exc)
        except PermissionError:
            error_msg = 'You do not have permission to rename %s' % id

        if error_msg and REQUEST is None:
            raise RenameError(error_msg)

        if REQUEST is not None:
            REQUEST.set('manage_tabs_message', error_msg)
            return self.manage_main(self, REQUEST, update_menu=1)

    def manage_cutObjects(self, ids, REQUEST=None):
        """Put a reference to the objects named in ids in the clipboard,
        marked for a cut operation."""
        if isinstance(ids, str):
            ids = [ids]
        oblist = []
        for id in ids:
            ob = self._safe_getOb(id)
            m = FileMoniker(ob)
            oblist.append(m.ids)
        cp = (1, oblist)
        cp = _cb_encode(cp)
        if REQUEST is not None:
            resp = REQUEST['RESPONSE']
            resp.setCookie('__lcp', cp, path='%s' % REQUEST['SCRIPT_NAME'])
            return self.manage_main(self, REQUEST, cb_dataValid=1)
        return cp

    def manage_copyObjects(self, ids, REQUEST=None, RESPONSE=None):
        """Put a reference to the objects named in ids in the clipboard,
        marked for a copy operation."""
        if isinstance(ids, str):
            ids = [ids]
        oblist = []
        for id in ids:
            ob = self._safe_getOb(id)
            m = FileMoniker(ob)
            oblist.append(m.ids)
        cp = (0, oblist)
        cp = _cb_encode(cp)
        if REQUEST is not None:
            resp = REQUEST['RESPONSE']
            resp.setCookie('__lcp', cp, path='%s' % REQUEST['SCRIPT_NAME'])
            return self.manage_main(self, REQUEST, cb_dataValid=1)
        return cp

    def manage_pasteObjects(self, cb_copy_data=None, REQUEST=None):
        """Paste objects from the clipboard into the current directory.
        The cb_copy_data parameter, if specified, should be the result
        of a previous call to manage_cutObjects or manage_copyObjects."""
        cp = None
        if cb_copy_data is not None:
            cp = cb_copy_data
        else:
            if REQUEST and '__lcp' in REQUEST:
                cp = REQUEST['__lcp']
        if cp is None:
            raise CopyError('No clipboard data found.')

        try:
            cp = _cb_decode(cp)
        except Exception as e:
            six.raise_from(CopyError('Clipboard Error'), e)

        oblist = []
        m = FileMoniker()
        op = cp[0]
        for ids in cp[1]:
            m.ids = ids
            try:
                ob = m.bind(self.root or self)
            except Exception:
                raise CopyError('Item Not Found')
            self._verifyObjectPaste(ob, REQUEST)
            oblist.append(ob)

        if op == 0:
            # Copy operation
            for ob in oblist:
                id = self._get_id(absattr(ob.id))
                self._copyOb(id, ob)

            if REQUEST is not None:
                return self.manage_main(self, REQUEST, update_menu=1,
                                        cb_dataValid=1)

        if op == 1:
            # Move operation
            for ob in oblist:
                id = absattr(ob.id)
                if not sanity_check(self, ob):
                    raise CopyError('This object cannot be pasted into itself')
                id = self._get_id(id)
                self._moveOb(id, ob)

            if REQUEST is not None:
                resp = REQUEST['RESPONSE']
                resp.setCookie('cp_', 'deleted',
                               path='%s' % REQUEST['SCRIPT_NAME'],
                               expires='Wed, 31-Dec-97 23:59:59 GMT')
                return self.manage_main(self, REQUEST, update_menu=1,
                                        cb_dataValid=0)
        return ''

    def cb_dataValid(self):
        """Return true if clipboard data seems valid."""
        try:
            _cb_decode(self.REQUEST['__lcp'])
            return 1
        except Exception:
            return 0

    def manage_delObjects(self, ids=[], REQUEST=None):
        """Delete files or subdirectories."""
        notfound = []
        if isinstance(ids, str):
            ids = [ids]
        if not ids:
            if REQUEST is not None:
                REQUEST.set('manage_tabs_message', 'No items specified')
                return self.manage_main(self, REQUEST)

        for id in ids:
            path = self._getpath(id)
            if not os.path.exists(path):
                notfound.append(id)
            else:
                self._delObject(id)

        if REQUEST is not None:
            if notfound:
                msg = 'Some items were not found: %s' % ', '.join(notfound)
            else:
                msg = 'Items deleted.'
            REQUEST.set('manage_tabs_message', msg)
            return self.manage_main(self, REQUEST, update_menu=1)

    def fileIds(self, spec=None):
        """Return a list of subobject ids.
        If 'spec' is specified, return only objects whose filename
        matches 'spec'."""
        if spec is None:
            spec = self.file_filter
        return self._ids(spec)

    def fileValues(self, spec=None, propagate=1):
        """Return a list of Local File objects.
        If 'spec' is specified, return only objects whose filename
        matches 'spec'."""
        if spec is None:
            spec = self.file_filter
        r = []
        a = r.append
        g = self._getfileob
        if propagate:
            for id in self._ids(spec):
                a(g(id, spec))
        else:
            for id in self._ids(spec):
                a(g(id))
        # sort that directories come first
        res = []
        for v in r:
            s = '%s%s' % ((v.type != 'directory'), v.id)
            res.append((s, v))

        return [x[1] for x in sorted(res)]

    def fileItems(self, spec=None, propagate=1):
        """Return a list of (id, fileobject) tuples.
        If 'spec' is specified, return only objects whose filename
        matches 'spec'
        """
        if spec is None:
            spec = self.file_filter
        r = []
        a = r.append
        g = self._getfileob
        if propagate:
            for id in self._ids(spec):
                a((id, g(id, spec)))
        else:
            for id in self._ids(spec):
                a((id, g(id)))
        return r

    def objectIds(self, spec=None):
        """Return a list of subobject ids.

        Returns a list of subobject ids of the current object.  If 'spec' is
        specified, returns objects whose meta_type matches 'spec'.
        """
        if self.catalog:
            return self._objectIds(spec)
        return ()

    def objectValues(self, spec=None):
        """Return a list of the actual subobjects.

        Returns a list of actual subobjects of the current object.  If
        'spec' is specified, returns only objects whose meta_type match 'spec'
        """
        if self.catalog:
            return self._objectValues(spec)
        return ()

    def objectItems(self, spec=None):
        """Return a list of (id, subobject) tuples.

        Returns a list of (id, subobject) tuples of the current object.
        If 'spec' is specified, returns only objects whose meta_type match
        'spec'
        """
        if self.catalog:
            return self._objectItems(spec)
        return ()

    def _objectIds(self, spec=None):
        g = self._safe_getOb
        ids = self._ids()
        if spec is not None:
            if isinstance(spec, str):
                spec = [spec]
            r = []
            a = r.append
            for id in ids:
                ob = g(id)
                if ob.meta_type in spec:
                    a(id)
            return r
        return ids

    def _objectValues(self, spec=None):
        r = []
        a = r.append
        g = self._safe_getOb
        if spec is not None:
            if isinstance(spec, str):
                spec = [spec]
            for id in self._ids():
                ob = g(id)
                if ob.meta_type in spec:
                    a(g(id))
        else:
            for id in self._ids():
                a(g(id))
        return r

    def _objectItems(self, spec=None):
        r = []
        a = r.append
        g = self._safe_getOb
        if spec is not None:
            if isinstance(spec, str):
                spec = [spec]
            for id in self._ids():
                ob = g(id)
                if ob.meta_type in spec:
                    a((id, g(id)))
        else:
            for id in self._ids():
                a((id, g(id)))
        return r

    def serverPath(self):
        """Return the full path of the directory object relative to the
        root of the server."""
        ids = []
        ob = self
        while 1:
            if not hasattr(ob, 'id'):
                break
            ids.append(absattr(ob.id))
            if not hasattr(ob, 'aq_parent'):
                break
            ob = ob.aq_parent
        ids.reverse()
        path_str = '/'.join(ids)
        return path_str.encode('UTF-8')

    def parentURL(self):
        """Return the URL of the parent directory."""
        url = self.REQUEST['URL2']
        spec = self.REQUEST.get('spec', None)
        if (spec is not None):
            if isinstance(spec, str):
                url = '%s?spec=%s' % (url, quote(spec))
            else:
                query = []
                for patt in spec:
                    query.append('spec:list=%s' % quote(patt))
                url = url + '?' + '&'.join(query)
        return url

    def defaultDocument(self):
        """Return the first default document found in this folder
        as a Zope object or None."""
        # Don't know why but self.default_document is sometimes empty
        try:
            files = self.default_document.split()
            for file in files:
                path = self._getpath(file)
                if (os.path.isfile(path)):
                    try:
                        return self._safe_getOb(file)
                    except Forbidden:
                        pass
        except Exception:
            pass
        return None

    def bobobase_modification_time(self):
        t = os.stat(self._local_path)[stat.ST_MTIME]
        return DateTime(t)

    #
    #  FTP and WebDAV Support - sbk
    #

    @security.private
    def PUT_factory(self, name, typ, body):
        """
        Dispatcher for PUT requests to non-existent IDs.
        """
        if name and (name.endswith('.pt') or name.endswith('.zpt')):
            from Products.PageTemplates.ZopePageTemplate import \
                ZopePageTemplate
            ob = ZopePageTemplate(name, body, content_type=typ)
        elif typ in ('text/html', 'text/xml', 'text/plain'):
            from OFS.DTMLDocument import DTMLDocument
            if not isinstance(body, (str, bytes)):
                body = body.read()
            ob = DTMLDocument(body, __name__=name)
        elif typ[:6] == 'image/':
            from OFS.Image import Image
            ob = Image(name, '', body, content_type=typ)
        else:
            from OFS.Image import File
            ob = File(name, '', body, content_type=typ)
        return ob

    if HAVE_FTP:

        def manage_FTPlist(self, REQUEST):
            """Directory listing for FTP"""
            out = ()
            files = list(self.fileItems())
            is_root = getattr(self,
                              'isTopLevelPrincipiaApplicationObject',
                              None)
            try:
                if not is_root and self.isTopLevelPrincipiaApplicationObject:
                    files.insert(0, ('..', self.aq_parent))
            except Exception:
                pass

            for k, v in files:
                try:
                    stat = marshal.loads(v.manage_FTPstat(REQUEST))
                except Exception:
                    stat = None
                if stat is not None:
                    out = out + ((k, stat),)
            return marshal.dumps(out)

        def manage_FTPstat(self, REQUEST):
            """Psuedo stat used for FTP listings"""
            # mode = 0040000 | 0770
            mode = 33272 | 0o0770
            mtime = self.bobobase_modification_time().timeTime()
            owner = group = 'Zope'
            return marshal.dumps((mode, 0, 0, 1, owner, group, 0, mtime, mtime,
                                  mtime))


class LocalFile(OFS.SimpleItem.Item, Acquisition.Implicit):

    """Object representing a file in the local file system."""

    meta_type = 'Local File'

    security = AccessControl.ClassSecurityInfo()

    security.declareProtected('FTP access', 'manage_FTPstat',  # NOQA
                              'manage_FTPget', 'manage_FTPlist', 'PUT',
                              'manage_FTPput')

    def __init__(self, parent, id, path, spec):
        """LocalFile __init__"""
        self.parent = parent
        self.id = id
        self.path = path
        self.type = self._getType()
        self.url = self._getURL(spec)
        self.plain_url = self._getPlainURL()
        self.icon = self._getIcon()
        self.size = self._getSize()
        self.mtime = self._getTime()
        self.display_size = self._getDisplaySize()
        self.display_mtime = self._getDisplayTime()

    def getObject(self):
        """Return a Zope object representing this local file."""
        return self.parent._safe_getOb(self.id)

    def get_size(self):
        """Return the size of the file."""
        return self.size

    def bobobase_modification_time(self):
        """bobobase_modification_time"""
        return self.mtime

    def _getURL(self, spec):
        """_getURL"""
        url = quote(self.id)
        if (self.type == 'directory') and (spec is not None):
            if isinstance(spec, str):
                return url + '?spec=' + quote(spec)
            else:
                query = []
                for patt in spec:
                    query.append('spec:list=%s' % quote(patt))
                return url + '?' + '&'.join(query)
        return url

    def _getPlainURL(self):
        """_getPlainURL"""
        return quote(self.id)

    def _getType(self):
        """Return the content type of a file."""
        name = self.id
        path = self.path
        if (os.path.isdir(path)):
            return 'directory'
        ext = os.path.splitext(name)[-1]
        t, c = _get_content_type(ext, self.parent._type_map)
        if t:
            return t
        try:
            f = open(path, 'rb')
            data = f.read(_test_read)
            f.close()
            ob = OFS.Image.File(name, name, data)
            _set_content_type(ob, t, data)
            return ob.content_type
        except EnvironmentError:
            return 'application/octet-stream'

    def _getIcon(self):
        """Return the path of the icon associated with this file type."""
        content_type = self.type.lower()
        _icon_map = self.parent._icon_map
        try:
            return _icon_map[content_type]
        except KeyError:
            pass
        content_type = content_type[:content_type.find('/')]
        try:
            return _icon_map[content_type]
        except KeyError:
            pass
        return _icon_base + 'generic.gif'

    def _getSize(self):
        """Return the size of the specified file or -1 if an error occurs.
        Return None if the path refers to a directory."""
        path = self.path
        if (os.path.isdir(path)):
            return 0
        try:
            return os.stat(path)[stat.ST_SIZE]
        except Exception:
            return -1

    def _getDisplaySize(self):
        """Return the size of a file or directory formatted for display."""
        size = self.size
        if size in (0, None):
            return '-' * 5
        if size == -1:
            return _unknown
        k = 1024.0
        if (size > k):
            size = size / k
            if (size > k):
                size = size / k
                return '%.1f MB' % size
            else:
                return '%.1f kB' % size
        else:
            return '%d bytes' % size

    def _getTime(self):
        """Return the last modified time of a file or directory
        or None if an error occurs."""
        try:
            return DateTime(os.stat(self.path)[stat.ST_MTIME])
        except Exception:
            pass

    def _getDisplayTime(self):
        """Return the last modified time of a file or directory formatted
        for display."""
        t = self.mtime
        if t is None:
            return _unknown
        return '%s %s' % (t.Time(), t.Date())

    #
    # FTP and WebDAV Support - sbk
    #

    def PUT(self, REQUEST, RESPONSE):
        """ Handle HTTP PUT requests """
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        self.write(REQUEST.get('BODY', ''))
        self.ZCacheable_invalidate()
        RESPONSE.setStatus(204)
        return RESPONSE

    manage_FTPput = PUT

    def manage_FTPget(self):
        """Get source for FTP download"""
        self.REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
        return self.read()

    getSize = get_size

    if HAVE_FTP:

        def manage_FTPstat(self, REQUEST):
            """Psuedo stat used for FTP listings"""
            size = self._getSize()
            # mode = 0100000 | 0660
            mode = 33152 | 0o0660
            if self.type == 'directory':
                size = 0
                # mode = 0040000 | 0770
                mode = 33272 | 0o0770
            mtime = self.bobobase_modification_time().timeTime()
            owner = group = 'Zope'
            return marshal.dumps((mode, 0, 0, 1, owner, group, size, mtime,
                                  mtime, mtime))


class FileMoniker:

    """A file moniker is a reference to an object in the file system."""

    def __init__(self, ob=None):
        """FileMoniker __init__"""
        if ob is None:
            return
        self.ids = []
        while 1:
            if not hasattr(ob, 'id'):
                break
            if ob.meta_type == 'Local File System':
                break
            self.ids.append(absattr(ob.id))
            ob = ob.aq_parent
        self.ids.reverse()

    def bind(self, root):
        """Return the file object named by this moniker"""
        ob = root
        for id in self.ids:
            ob = ob._safe_getOb(id)
        return ob


class LocalFS(LocalDirectory, OFS.PropertyManager.PropertyManager,
              Persistence.Persistent, RoleManager):

    """Object that creates Zope objects from files in the local file system."""

    meta_type = 'Local File System'
    zmi_icon = 'far fa-hdd text-danger'

    manage_options = (
        ({'label': 'Contents', 'action': 'manage_main'},
         {'label': 'View', 'action': ''},
         {'label': 'Properties', 'action': 'manage_propertiesForm'},
         {'label': 'Security', 'action': 'manage_access'},
         {'label': 'Upload', 'action': 'manage_uploadForm'})
    )

    __ac_permissions__ = (
        ('View', ('',)),
        ('View management screens',
            ('manage', 'manage_main')),
        ('Change Local File System properties',
            ('manage_propertiesForm', 'manage_changeProperties')),
        ('Access contents information',
            ('fileIds', 'fileValues', 'fileItems')),
        ('Upload local files',
            ('manage_uploadForm', 'manage_upload')),
        ('Overwrite local files', ('manage_overwrite',)),
        ('Manage local files',
            ('manage_cutObjects', 'manage_copyObjects', 'manage_pasteObjects',
             'manage_renameForm', 'manage_renameObject',
             'manage_createDirectory')),
        ('Delete local files', ('manage_delObjects',)),
        )

    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
        {'id': 'basepath', 'type': 'string', 'mode': 'w'},
    )
    if (_iswin32):
        _properties = _properties + (
            {'id': 'username', 'type': 'string', 'mode': 'w'},
            {'id': 'password', 'type': 'string', 'mode': 'w'},
            )
    _properties = _properties + (
        {'id': 'default_document', 'type': 'string', 'mode': 'w'},
        {'id': 'type_map', 'type': 'lines', 'mode': 'w'},
        {'id': 'icon_map', 'type': 'lines', 'mode': 'w'},
        {'id': 'catalog', 'type': 'boolean', 'mode': 'w'},
        {'id': 'file_filter', 'type': 'string', 'mode': 'w'},
    )

    default_document = 'index.html default.html'
    username = _share = ''
    _connected = 0
    isPrincipiaFolderish = 1  # leaving this at one though
    catalog = 0
    root = None
    password = _password = ''

    _type_map = _types
    _icon_map = _icons
    type_map = _typemap2list(_types)
    icon_map = _iconmap2list(_icons)
    file_filter = None

    def __init__(self, id, title, basepath, username, password):
        """LocalFS __init__"""
        LocalDirectory.__init__(self, id, basepath, self, self.catalog,
                                self._type_map, self._icon_map,
                                self.file_filter)
        self.title = title

        self.basepath = self._local_path = basepath
        if (_iswin32):
            self.username = username
            self._password = password or ''
            m = unc_expr.match(self.basepath)
            if (m is not None) and (self.username):
                self._share = m.group(1)
                self._connect()
            else:
                self._share = ''

    def manage_editProperties(self, REQUEST):
        """Edit object properties via the web.
        The purpose of this method is to change all property values,
        even those not listed in REQUEST; otherwise checkboxes that
        get turned off will be ignored.  Use manage_changeProperties()
        instead for most situations.
        """
        type_map = self.type_map
        icon_map = self.icon_map
        username = self.username
        password = self.password

        super(LocalFS, self).manage_editProperties(REQUEST)

        if not self.file_filter.strip():
            self.file_filter = None
        if self.file_filter == 'None':
            self.file_filter = None
        if self.type_map != type_map:
            self._type_map = _list2typemap(self.type_map)
        if self.icon_map != icon_map:
            self._icon_map = _list2iconmap(self.icon_map)
        if (_iswin32):
            if self.username != username or self.password != password:
                self._password = self.password
                if (self._connected):
                    self._disconnect()
                m = unc_expr.match(self.basepath)
                if (m is not None) and (self.username):
                    self._share = m.group(1)
                    self._connect()
                else:
                    self._share = ''
            self.password = ''
        message = 'Saved changes.'
        return self.manage_propertiesForm(self, REQUEST,
                                          manage_tabs_message=message,
                                          update_menu=1)

    def manage_changeProperties(self, REQUEST=None, **kw):
        """Change existing object properties.

        Change object properties by passing either a mapping object
        of name:value pairs {'foo':6} or passing name = value parameters
        """
        type_map = self.type_map
        icon_map = self.icon_map
        username = self.username
        password = self._password

        super(LocalFS, self).manage_changeProperties(REQUEST=REQUEST, **kw)

        if self.type_map != type_map:
            self._type_map = _list2typemap(self.type_map)
        if self.icon_map != icon_map:
            self._icon_map = _list2iconmap(self.icon_map)
        if (_iswin32):
            if self.username != username or self.password != password:
                self._password = self.password
                if (self._connected):
                    self._disconnect()
                m = unc_expr.match(self.basepath)
                if (m is not None) and (self.username):
                    self._share = m.group(1)
                    self._connect()
                else:
                    self._share = ''
            self.password = ''

    def _connect(self):
        """_connect"""
        win32wnet.WNetAddConnection2(1, None, self._share, None,
                                     self.username or None,
                                     self._password or None)
        self._connected = 1

    def _disconnect(self):
        """_disconnect"""
        win32wnet.WNetCancelConnection2(self._share, 0, 0)
        self._connected = 0

    def _check_connected(self):
        """_check_connected"""
        if (self._share and not self._connected):
            self._connect()

    def _ids(self, spec=None):
        """_ids"""
        self._check_connected()
        return LocalDirectory._ids(self, spec)

    def _getfileob(self, id, spec=None):
        """_getfileob"""
        self._check_connected()
        return LocalDirectory._getfileob(self, id, spec)

    def _getOb(self, id, default=_marker):
        """_getOb"""
        self._check_connected()
        return LocalDirectory._getOb(self, id, default)

    def bobobase_modification_time(self):
        """bobobase_modification_time"""
        return Persistence.Persistent.bobobase_modification_time(self)

    def hasDefaultDocument(self):
        """Return true if is Directory and has default doc"""
        # Don't know why but self.default_document is sometimes empty
        try:
            files = self.default_document.split()
            for file in files:
                path = self._getpath(file)
                if (os.path.isfile(path)):
                    try:
                        return self._safe_getOb(file)
                    except Forbidden:
                        pass
        except Exception:
            pass
        return None


def manage_addLocalFS(self, id, title, basepath, username=None,
                      password=None, REQUEST=None):
    """Add a local file system object to a folder

    In addition to the standard Zope object-creation arguments,
    'id' and 'title', the following arguments are defined:

        basepath -- The base path of the local files.
        username -- Username for a network share (win32 only).
        password -- Password for a network share (win32 only).
    """

    ob = LocalFS(id, title, basepath, username, password)
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)


manage_addLocalFSForm = HTMLFile('dtml/methodAdd', globals())



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Tests for the LocalDirectory class

import os
import tempfile
import unittest

from Acquisition import aq_base
from DateTime.DateTime import DateTime
from Products.PageTemplates.ZopePageTemplate import manage_addPageTemplate
from Testing.makerequest import makerequest
from zExceptions import BadRequest

from .helpers import ADMIN_USER
from .helpers import LOCALFS_ROOT
from .helpers import FilesystemTestSupport
from .helpers import FunctionalTestCase


class LocalDirectoryAndFileTests(unittest.TestCase, FilesystemTestSupport):

    def _getTargetClass(self):
        # Using the LocalFS class, it's a subclass of LocalDirectory
        # that does not override any of its methods and is easier to initialize
        from ..LocalFS import LocalFS

        return LocalFS

    def _makeOne(self, *args):
        return self._getTargetClass()(*args)

    def _makeSimple(self):
        klass = self._getTargetClass()
        lfs = klass('localfs', 'LocalFS Title', LOCALFS_ROOT, 'user', 'pw')
        lfs.catalog = True
        return lfs

    def _makeLocalFile(self):
        lfs = self._makeSimple()
        manage_addPageTemplate(lfs, 'test.pt', b'<html>')
        lfs.manage_createDirectory('folder')
        return (lfs, lfs._getfileob('test.pt'), lfs._getfileob('folder'))

    def tearDown(self):
        self.cleanup_files()
        super(LocalDirectoryAndFileTests, self).tearDown()

    def test_instantiation(self):
        lfs = self._makeSimple()
        self.assertEqual(lfs.getId(), 'localfs')
        self.assertEqual(lfs.basepath, LOCALFS_ROOT)
        self.assertIs(aq_base(lfs.root), aq_base(lfs))
        self.assertTrue(lfs._type_map)
        self.assertTrue(lfs._icon_map)
        self.assertIsNone(lfs.file_filter)

    def test___bobo_traverse__PUT(self):
        lfs = self._makeSimple()
        req = {'REQUEST_METHOD': 'PUT'}

        returned = lfs.__bobo_traverse__(req, 'foo')
        self.assertEqual(returned.getId(), 'foo')
        self.assertEqual(returned.title, "LockNull Resource 'foo'")

    def test___getitem__invalid_input(self):
        lfs = self._makeSimple()

        with self.assertRaises(TypeError) as context:
            lfs[None]
        self.assertEqual(str(context.exception), 'index must be a string')

    def test__getpath(self):
        lfs = self._makeSimple()
        self.assertEqual(lfs._getpath('foo'),
                         os.path.join(LOCALFS_ROOT, 'foo'))

    def test__checkId(self):
        lfs = self._makeSimple()

        # No return value signals success
        self.assertIsNone(lfs._checkId('foo'))

        with self.assertRaises(BadRequest) as context:
            lfs._checkId('')
        self.assertIn('No id was specified', str(context.exception))

        with self.assertRaises(BadRequest) as context:
            lfs._checkId('foo&bar')
        self.assertIn('characters illegal in filenames',
                      str(context.exception))

        with self.assertRaises(BadRequest) as context:
            lfs._checkId('_secret')
        self.assertIn('it begins with an underscore', str(context.exception))

        with self.assertRaises(BadRequest) as context:
            lfs._checkId('.gitkeep', allow_dup=False)
        self.assertIn('it is already in use', str(context.exception))
        self.assertIsNone(lfs._checkId('.gitkeep', allow_dup=True))

    def test_PageTemplate(self):
        lfs = self._makeSimple()

        self.assertNotFileExists('test.pt')
        self.assertNotIn('test.pt', lfs.objectIds())

        manage_addPageTemplate(lfs, 'test.pt', '', b'<html>')
        self.assertFileExists('test.pt')
        self.assertIn('test.pt', lfs.objectIds())

        pt_obj = lfs['test.pt']
        self.assertEqual(pt_obj.data, b'<html>')

    def test_cut_copy_paste_rename_delete(self):
        lfs = self._makeSimple()

        lfs.manage_createDirectory('folder')
        manage_addPageTemplate(lfs, 'test.pt', '', b'<html>')
        self.assertIn('test.pt', lfs.objectIds())
        self.assertNotIn('test.pt', lfs.folder.objectIds())

        ref = lfs.manage_copyObjects('test.pt')
        lfs.folder.manage_pasteObjects(ref)
        self.assertIn('test.pt', lfs.objectIds())
        self.assertIn('test.pt', lfs.folder.objectIds())

        self.assertNotIn('renamed_test.pt', lfs.objectIds())
        lfs.manage_renameObject('test.pt', 'renamed_test.pt')
        self.assertNotIn('test.pt', lfs.objectIds())
        self.assertIn('renamed_test.pt', lfs.objectIds())

        self.assertNotIn('renamed_test.pt', lfs.folder.objectIds())
        ref = lfs.manage_cutObjects('renamed_test.pt')
        lfs.folder.manage_pasteObjects(ref)
        self.assertNotIn('renamed_test.pt', lfs.objectIds())
        self.assertIn('renamed_test.pt', lfs.folder.objectIds())

        lfs.folder.manage_delObjects('renamed_test.pt')
        self.assertNotIn('renamed_test.pt', lfs.folder.objectIds())

    def test_objectListMethods(self):
        lfs = self._makeSimple()

        # If the "catalog" variable is false these methods won't work
        lfs.catalog = False
        self.assertFalse(lfs.objectIds())
        self.assertFalse(lfs.objectIds(spec=['Local Directory']))
        self.assertFalse(lfs.objectIds(spec='Local Directory'))
        self.assertFalse(lfs.objectValues())
        self.assertFalse(lfs.objectValues(spec=['Local Directory']))
        self.assertFalse(lfs.objectValues(spec='Local Directory'))
        self.assertFalse(lfs.objectItems())
        self.assertFalse(lfs.objectItems(spec=['Local Directory']))
        self.assertFalse(lfs.objectItems(spec='Local Directory'))

        lfs.catalog = True
        self.assertEqual(lfs.objectIds(), ['.gitkeep'])
        self.assertFalse(lfs.objectIds(spec=['Local Directory']))
        self.assertFalse(lfs.objectIds(spec='Local Directory'))
        self.assertEqual(len(lfs.objectValues()), 1)
        self.assertFalse(lfs.objectValues(spec=['Local Directory']))
        self.assertFalse(lfs.objectValues(spec='Local Directory'))
        self.assertEqual(len(lfs.objectItems()), 1)
        self.assertFalse(lfs.objectItems(spec=['Local Directory']))
        self.assertFalse(lfs.objectItems(spec='Local Directory'))

        lfs.manage_createDirectory('foo')
        self.assertEqual(lfs.objectIds(), ['.gitkeep', 'foo'])
        self.assertEqual(lfs.objectIds(spec=['Local Directory']), ['foo'])
        self.assertEqual(lfs.objectIds(spec='Local Directory'), ['foo'])
        self.assertEqual(len(lfs.objectValues()), 2)
        self.assertEqual(len(lfs.objectValues(spec=['Local Directory'])), 1)
        self.assertEqual(len(lfs.objectValues(spec='Local Directory')), 1)
        self.assertEqual(len(lfs.objectItems()), 2)
        self.assertEqual(len(lfs.objectItems(spec=['Local Directory'])), 1)
        self.assertEqual(len(lfs.objectItems(spec='Local Directory')), 1)

    def test_fileMethods(self):
        lfs = self._makeSimple()

        self.assertEqual(lfs.fileIds(), ['.gitkeep'])
        self.assertEqual(len(lfs.fileValues()), 1)
        self.assertEqual(len(lfs.fileValues(propagate=False)), 1)
        self.assertEqual(len(lfs.fileItems()), 1)
        self.assertEqual(len(lfs.fileItems(propagate=False)), 1)

        self.assertFalse(lfs.fileIds(spec=['Local Directory']))
        self.assertFalse(lfs.fileValues(spec=['Local Directory']))
        self.assertFalse(lfs.fileItems(spec=['Local Directory']))

    def test_defaultDocument(self):
        lfs = self._makeSimple()
        lfs.default_document = 'index.html default.html'

        self.assertIsNone(lfs.defaultDocument())

        manage_addPageTemplate(lfs, 'default.html', b'<html>')
        doc = lfs.defaultDocument()
        self.assertEqual(doc.getId(), 'default.html')

        manage_addPageTemplate(lfs, 'index.html', b'<html>')
        doc = lfs.defaultDocument()
        self.assertEqual(doc.getId(), 'index.html')

    def test_bobobase_modification_time(self):
        lfs = self._makeSimple()
        manage_addPageTemplate(lfs, 'test.pt', '', b'<html>')

        self.assertIsInstance(lfs['test.pt'].bobobase_modification_time(),
                              DateTime)

    # LocalFile-specific tests

    def test_getObject(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertEqual(lof.meta_type, 'Local File')
        self.assertEqual(lof.getObject().meta_type, 'File')

    def test_get_size(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertTrue(lof.get_size())
        self.assertFalse(lod.get_size())

    def test_bobobase_modification_time_localfile(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertIsInstance(lof.bobobase_modification_time(), DateTime)

    def test__getURL(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertEqual(lof._getURL('Local File'), 'test.pt')
        self.assertEqual(lod._getURL('Local File'), 'folder?spec=Local%20File')
        self.assertEqual(
            lod._getURL(['Local File', 'Local Directory']),
            'folder?spec:list=Local%20File&spec:list=Local%20Directory')

    def test__getType(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertEqual(lof._getType(), 'text/html')
        self.assertEqual(lod._getType(), 'directory')

    def test__getIcon(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertEqual(lof._getIcon(), 'misc_/LocalFS/layout.gif')
        self.assertEqual(lod._getIcon(), 'misc_/LocalFS/dir.gif')

        # an unknown item
        lof.type = 'unknown'
        self.assertEqual(lof._getIcon(), 'misc_/LocalFS/generic.gif')

    def test__getDisplaySize(self):
        lfs, lof, lod = self._makeLocalFile()

        self.assertIn('bytes', lof._getDisplaySize())
        self.assertEqual(lod._getDisplaySize(), '-----')

        # large size
        lof.size = 8543
        self.assertEqual(lof._getDisplaySize(), '8.3 kB')

        lof.size = 678358764
        self.assertEqual(lof._getDisplaySize(), '646.9 MB')

        # invalid size
        lof.size = -1
        self.assertEqual(lof._getDisplaySize(), '(unknown)')


class LocalDirectoryFunctionalTests(FunctionalTestCase, FilesystemTestSupport):

    def tearDown(self):
        self.cleanup_files()
        super(LocalDirectoryFunctionalTests, self).tearDown()

    def test_manage_upload(self):
        self.login(ADMIN_USER)
        lfs = makerequest(self.folder.localfs)
        manage_addPageTemplate(lfs, 'test.pt', '', b'<html>')

        # existing file
        with tempfile.NamedTemporaryFile(dir=LOCALFS_ROOT, delete=True) as fp:
            fp.write(b'<html></html>')
            fp.seek(0)
            # existing file
            lfs.manage_upload(fp, 'test.pt')
        self.assertEqual(lfs['test.pt'].data, b'<html></html>')

        # new file
        self.assertNotIn('new_test.pt', lfs.objectIds())
        with tempfile.NamedTemporaryFile(dir=LOCALFS_ROOT, delete=True) as fp:
            fp.write(b'<html></html>')
            fp.seek(0)
            lfs.manage_upload(fp, 'new_test.pt')
        self.assertIn('new_test.pt', lfs.objectIds())
        self.assertEqual(lfs['new_test.pt'].data, b'<html></html>')

    def test_Directory(self):
        self.login(ADMIN_USER)
        lfs = makerequest(self.folder.localfs)
        req = lfs.REQUEST

        self.assertNotFolderExists('foo')
        self.assertIsNone(lfs.__bobo_traverse__(req, 'foo'))
        with self.assertRaises(AttributeError) as context:
            getattr(lfs, 'foo')
        self.assertIn('foo', str(context.exception))
        with self.assertRaises(AttributeError) as context:  # MISBEHAVIOR!
            lfs['foo']
        self.assertEqual(str(context.exception), 'foo')
        self.assertFalse(lfs.objectIds(spec=['Local Directory']))

        returned = lfs.manage_createDirectory('foo', REQUEST=req)
        self.assertIn('Directory foo has been created', returned)
        self.assertFolderExists('foo')
        self.assertIn('foo', lfs.objectIds())
        self.assertEqual(lfs.objectIds(spec=['Local Directory']), ['foo'])

        # Calling manage_createDirectory again won't do anything
        returned = lfs.manage_createDirectory('foo', REQUEST=req)
        self.assertIn('Directory foo already exists', returned)

        lfs.manage_delObjects(ids=['foo'])
        self.assertNotFolderExists('foo')
        self.assertFalse(lfs.objectIds(spec=['Local Directory']))



__version__ = '2.0b1'

import os

from setuptools import find_packages
from setuptools import setup


NAME = 'Products.LocalFS'
here = os.path.abspath(os.path.dirname(__file__))
_boundary = '\n' + ('-' * 60) + '\n\n'


def _read(name):
    f = open(os.path.join(here, name))
    return f.read()


setup(name=NAME,
      version=__version__,
      license='BSD License',
      description='The Local File System product',
      long_description=_read('README.rst'),
      classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Web Environment',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Framework :: Zope :: 5',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
        ],
      keywords='web application server zope zope2',
      author='Jonothan Farr and contributors',
      author_email='jfarr@speakeasy.org',
      maintainer='Jens Vagelpohl',
      maintainer_email='jens@plyp.com',
      url='https://github.com/dataflake/%s' % NAME,
      project_urls={
          'Issue Tracker': 'https://github.com/dataflake/%s/issues' % NAME,
          'Sources': 'https://github.com/dataflake/%s' % NAME,
      },
      packages=find_packages('src'),
      include_package_data=True,
      namespace_packages=['Products'],
      package_dir={'': 'src'},
      zip_safe=False,
      python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      install_requires=[
        'setuptools',
        'Zope',
        'Products.PythonScripts',
        ],
      extras_require={
          'docs': ['Sphinx < 2;python_version < "3"',
                   'Sphinx;python_version >= "3"']
        },
      )



##############################################################################
#
# Copyright (c) 1999 Jonothan Farr and contributors
# All rights reserved. Written by Jonothan Farr <jfarr@speakeasy.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# In accordance with the license provided for by the software upon
# which some of the source code has been derived or used, the following
# acknowledgement is hereby provided:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
##############################################################################
# Helper functions and definitions for LocalFS tests

import os
import shutil

from Testing.ZopeTestCase import ZopeTestCase

from ..LocalFS import manage_addLocalFS


TESTS_PATH = os.path.dirname(os.path.abspath(__file__))
LOCALFS_ROOT = os.path.join(TESTS_PATH, 'files')
ADMIN_USER = 'admin'


class FunctionalTestCase(ZopeTestCase):

    _setup_fixture = True

    def _setup(self):
        super(FunctionalTestCase, self)._setup()

        # Add an admin user
        uf = self.folder.acl_users
        uf.userFolderAddUser(ADMIN_USER, 'foobar', ['Manager'], [])

        # Add a LocalFS instance
        manage_addLocalFS(self.folder,
                          'localfs',
                          'LocalFS Title',
                          LOCALFS_ROOT)

        # Make sure the objectXXX methods work
        self.folder.localfs.tree_view = True
        self.folder.localfs.catalog = True


class FilesystemTestSupport:
    """ Mix-in with utility methods for manipulating filesystem files """

    def cleanup_files(self):
        filenames = self.list_folder()
        filenames.remove('.gitkeep')

        for filename in filenames:
            path = os.path.join(LOCALFS_ROOT, filename)
            if os.path.isfile(path):
                os.unlink(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)

    def list_folder(self, subpath=None):
        if subpath:
            path = os.path.join(LOCALFS_ROOT, subpath)
        else:
            path = LOCALFS_ROOT
        return os.listdir(path)

    def assertFolderExists(self, subpath):
        path = os.path.join(LOCALFS_ROOT, subpath)
        if not os.path.isdir(path):
            raise AssertionError('Folder does not exist: %s' % path)

    def assertNotFolderExists(self, subpath):
        path = os.path.join(LOCALFS_ROOT, subpath)
        if os.path.isdir(path):
            raise AssertionError('Folder exists: %s' % path)

    def assertFileExists(self, subpath):
        path = os.path.join(LOCALFS_ROOT, subpath)
        if not os.path.isfile(path):
            raise AssertionError('File does not exist: %s' % path)

    def assertNotFileExists(self, subpath):
        path = os.path.join(LOCALFS_ROOT, subpath)
        if os.path.isfile(path):
            raise AssertionError('File exists: %s' % path)

    def read_file(self, subpath):
        with open(os.path.join(LOCALFS_ROOT, subpath), 'rb') as fp:
            return fp.read()


class FakeContent:

    basepath = ''
    _local_path = ''
